#include "header\GBA.h"
#include "header\BG.h"
#include "header\Timer.h"
#include "header\DMA.h"
#include "header\IRQ.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\Blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "data.h"
#include "resource\BitMapFont.h"

#include <stdlib.h>
#include <string.h>

//const int __gba_multiboot;	//GBANp

#define TRUE  (1)
#define FALSE (0)

#define UNIT_MAX_CNT 8			//R}̐i3`8܂Łj

#define LINE_X       10			//\Jnʒu
#define LINE_Y       10
#define LINE_LENGTH 120			//A~_̏c̐
#define LINE_WIDTH   30			//A~_̉̐

#define UNIT_STATE_FRONT 1		//R}̏Ԃ\
#define UNIT_STATE_LEFT  2
#define UNIT_STATE_RIGHT 3

#define SPRITE_REGIST_START_NUMBER 512

#define COLOR_WHITE RGB(31,31,31)
#define COLOR_BLACK RGB( 0, 0, 0)

//---------------------------------------------------------------------------
//A~_̏ɑ炷R}
typedef struct {
	u8  regFlag;			//^XN̓o^(TRUE or FALSE)
	u8  state;				//R}̏
	u8  x;
	u8  y;
	u8  cx;
	u8  cy;
	u16 SprNo;
	u16 SprTile;
	void (*pAction)(u8);	//ړ̊֐
	void (*pDestroy)(u8);	//I̊֐
}ST_UNIT;

ST_UNIT UNIT[UNIT_MAX_CNT];

u8 TaskRegistCnt;	//^XN̓o^
//---------------------------------------------------------------------------
void Init();

void TaskInit();
void TaskStart();
void TaskStop();
void TaskExec();

void TaskUnitAction(u8);
void TaskUnitDestroy(u8);

void DrawTateLine();
void DrawYokoLine();

void DispAllUnit();

void IRQ_USER_Handler();
void IRQ_USER_Setup();
//---------------------------------------------------------------------------
int main(void)
{
ReStart:

	Init();
	TaskInit();
	SetMode(MODE_3 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_1D);

	Mode3DrawBitMapStr(1,18, "Push A Button");
	WaitKey(KEY_A);		//SrandĂ܂

	DrawTateLine();
	DrawYokoLine();
	DispAllUnit();

	Mode3DrawBitMapStr(1,18, "Push B Button");
	WaitKey(KEY_B);		//[CEEE
	Mode3DrawBitMapStr(1,18, "             ");

	TaskStart();		//hI

	for(;;)
	{
		WaitForVsync();
		
		if(TaskRegistCnt <= 0)
		{
			break;
		}
	}

	TaskStop();

	Mode3DrawBitMapStr(1,18, "Push B Button");
	WaitKey(KEY_B);

	goto ReStart;

	return 0;
}
//---------------------------------------------------------------------------
void Init()
{
	Mode3DrawCls(COLOR_BLACK);
	Mode3SetBitMapFont((u8*)BitMapFont, 8,8,0);

	SpriteInitialize();
	SpriteSetPal16((u16*)EtfsGetDataOffset(_SPRCHR_IMG_PAL_OFFSET));
	SpriteSetData((u16*)EtfsGetDataOffset(_SPRCHR_IMG_BIN_OFFSET), _SPRCHR_IMG_BIN_SIZE);
}
//---------------------------------------------------------------------------
void TaskInit()
{
	u8 i;

	TaskRegistCnt = UNIT_MAX_CNT;

	for(i=0; i<UNIT_MAX_CNT ;i++)
	{
		UNIT[i].regFlag   = TRUE;
		UNIT[i].pAction   = &TaskUnitAction;
		UNIT[i].pDestroy  = &TaskUnitDestroy;
		
		UNIT[i].state     = UNIT_STATE_FRONT;
		UNIT[i].x         = LINE_X + LINE_WIDTH * i;
		UNIT[i].y         = LINE_Y + 120;
		UNIT[i].cx        = 8;
		UNIT[i].cy        = 8;
		UNIT[i].SprNo     = i;
		UNIT[i].SprTile   = SPRITE_REGIST_START_NUMBER + i*2;

		SpriteSetSize(UNIT[i].SprNo, SP_SIZE_8,SP_SQUARE,SP_COLOR_16);
		SpriteSetChr(UNIT[i].SprNo, UNIT[i].SprTile);
	}

	//荞
	IRQ_USER_Setup();
}
//---------------------------------------------------------------------------
void DrawTateLine()
{
	u8 i;

	for(i=0; i<UNIT_MAX_CNT; i++)
	{
		Mode3DrawLine(LINE_X+(LINE_WIDTH*i),
					  LINE_Y,
					  LINE_X+(LINE_WIDTH*i),
					  LINE_Y+LINE_LENGTH,
					  COLOR_WHITE);
	}
}
//---------------------------------------------------------------------------
void DrawYokoLine()
{
	u8 y, rnd;
	
	for(y=5; y<LINE_LENGTH; y+=5)
	{
		rnd = rand() % (UNIT_MAX_CNT-1);

		Mode3DrawLine(LINE_X+(rnd+0)*LINE_WIDTH,
					  LINE_Y+y,
					  LINE_X+(rnd+1)*LINE_WIDTH,
					  LINE_Y+y,
					  COLOR_WHITE);
	}
}
//---------------------------------------------------------------------------
void DispAllUnit()
{
	u8 i;

	for(i=0; i<UNIT_MAX_CNT; i++)
	{
		SpriteMove(UNIT[i].SprNo, UNIT[i].x, UNIT[i].y);
	}
}
//---------------------------------------------------------------------------
void TaskExec()
{
	u8 i;
	
	for(i=0; i<UNIT_MAX_CNT; i++)
	{
		if(UNIT[i].regFlag == TRUE)
		{
			UNIT[i].pAction(i);		//ړ
			UNIT[i].pDestroy(i);	//^XNI
		}
	}
}
//---------------------------------------------------------------------------
void IRQ_USER_Setup(){
	REG_IME = IRQ_MASTER_OFF;

	REG_INTERUPT  = (u32)IRQ_USER_Handler;
	REG_IE       |= IRQ_BIT_TIMER0;

	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void IRQ_USER_Handler(){
	REG_IME  = IRQ_MASTER_OFF;
	u16 Flag = REG_IF;

	if(Flag & IRQ_BIT_TIMER0)
	{
		TaskExec();
		DispAllUnit();

		REG_TM0D   = 0xF000;
	}

	REG_IF  = Flag;
	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void TaskStart()
{
	REG_TM0D   = 0xF000;
	REG_TM0CNT = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void TaskStop()
{
	REG_TM0D   = 0x0000;
	REG_TM0CNT = TM_DISABLE;
}
//---------------------------------------------------------------------------
void TaskUnitAction(u8 i)
{
	u16  LeftColor  = Mode3GetDot(UNIT[i].x-1, UNIT[i].y);
	u16  RightColor = Mode3GetDot(UNIT[i].x+1, UNIT[i].y);
	
	if(LeftColor == COLOR_WHITE
		&& UNIT[i].state != UNIT_STATE_RIGHT)
	{
		UNIT[i].x--;
		UNIT[i].state = UNIT_STATE_LEFT;
	}
	else if(RightColor == COLOR_WHITE
		&& UNIT[i].state != UNIT_STATE_LEFT)
	{
		UNIT[i].x++;
		UNIT[i].state = UNIT_STATE_RIGHT;
	}
	else
	{
		UNIT[i].y--;
		UNIT[i].state = UNIT_STATE_FRONT;
	}
}
//---------------------------------------------------------------------------
void TaskUnitDestroy(u8 i)
{
	if(UNIT[i].y == LINE_Y)
	{
		UNIT[i].regFlag = FALSE;
		TaskRegistCnt--;
	}
}
